@@ -10,31 +10,24 @@
 
 import requests
 import urllib3
-import codecs
-import json, os, glob
+import json
+import os
 from user_agent import generate_user_agent
-from typing import Union
-from OpaExceptions.OpaExceptions import (
-    CheckPermissionError,
-    ConnectionsError,
-    DeleteDataError,
-    DeletePolicyError,
-    PathNotFoundError,
-    PolicyNotFoundError,
-    RegoParseError,
-    SSLError,
-    FileError,
-    TypeExecption
-)
+from typing import Union, Dict
+from urllib.parse import urlencode
+from .OpaExceptions import (CheckPermissionError,
+                            ConnectionsError, DeleteDataError,
+                            DeletePolicyError, PathNotFoundError,
+                            PolicyNotFoundError, RegoParseError,
+                            SSLError, FileError, TypeExecption,
+                            QueryExecuteError)
 
-json_reader = codecs.getreader("utf-8")
-
-
-__version__ = "1.0.6"
+__version__ = "1.3.2"
+__author__ = "Tural Muradov"
+__license__ = 'MIT'
 
 
 class OpaClient:
-
     """ OpaClient client object to connect and manipulate OPA service.
         ```
         The class object holds session information necesseary to connect OPA service.
@@ -55,12 +48,12 @@ class OpaClient:
 
     def __init__(
         self,
-        host: str="localhost",
-        port: int=8181,
-        version: str="v1",
-        ssl: bool=False,
-        cert:Union[None,str]=None,
-        headers: Union[None,dict]=None,
+        host: str = "localhost",
+        port: int = 8181,
+        version: str = "v1",
+        ssl: bool = False,
+        cert: Union[None, str] = None,
+        headers: Union[None, dict] = None,
         **kwargs,
     ):
         host = host.lstrip()
@@ -70,25 +63,25 @@ class OpaClient:
         self.__data_root = "{}/data/{}"
         self.__secure = False
         self.__schema = "http://"
+        self.retries = kwargs.get("retries", 2)
+        self.timeout = kwargs.get("timeout", 1.5)
 
-        if type(port) not in [int]:
+        if not isinstance(self.__port, int):
             raise TypeError("The port must be integer")
-        
+
         if ssl:
             self.__ssl = ssl
             self.__cert = cert
             self.__secure = True
             self.__schema = "https://"
 
-        if not cert:
-            if ssl:
-                raise SSLError("ssl=True", "Make sure you  provide cert file")
+        if not cert and ssl is True:
+            raise SSLError("ssl=True", "Make sure you  provide cert file")
 
         if host.startswith("https://"):
             self.__host = host
-            self.__root_url = "{}:{}/{}".format(
-                self.__host, self.__port, self.__version
-            )
+            self.__root_url = "{}:{}/{}".format(self.__host, self.__port,
+                                                self.__version)
 
         elif host.startswith("http://"):
             self.__host = host
@@ -98,16 +91,14 @@ class OpaClient:
                     "With ssl enabled not possible to have connection with http",
                 )
 
-            self.__root_url = "{}:{}/{}".format(
-                self.__host, self.__port, self.__version
-            )
+            self.__root_url = "{}:{}/{}".format(self.__host, self.__port,
+                                                self.__version)
 
         else:
             self.__host = host
 
-            self.__root_url = "{}{}:{}/{}".format(
-                self.__schema, self.__host, self.__port, self.__version
-            )
+            self.__root_url = "{}{}:{}/{}".format(self.__schema, self.__host,
+                                                  self.__port, self.__version)
 
         if headers:
             self.__headers = requests.utils.default_headers()
@@ -126,9 +117,7 @@ class OpaClient:
             )
             self.__session = self.__manager.request
         else:
-            self.__manager = urllib3.PoolManager(
-                headers=self.__headers
-            )
+            self.__manager = urllib3.PoolManager(headers=self.__headers)
             self.__session = self.__manager.request
 
     def __del__(self):
@@ -138,27 +127,60 @@ class OpaClient:
         """
         Close all currently open connections to the OPA server
         """
-        self.__manager.clear()
+        try:
+            self.__manager.clear()
+        except:  # noqa: E722
+            pass
 
     def check_connection(self):
         """
-            Checks whether established connection config True or not.   
+            Checks whether established connection config True or not.
             if not properly configured will raise an ConnectionError.
         """
-        
+
         url = self.__policy_root.format(self.__root_url, "")
         try:
-            response = self.__session("GET", url, retries=2, timeout=1.5)
+            response = self.__session(
+                "GET", url, retries=self.retries, timeout=self.timeout)
             if response.status == 200:
                 return "Yes I'm here :)"
 
         except Exception:
-            raise ConnectionsError("service unreachable", "check config and try again")
+            raise ConnectionsError("service unreachable",
+                                   "check config and try again")
 
-        raise ConnectionsError("service unreachable", "check config and try again")
+        raise ConnectionsError("service unreachable",
+                               "check config and try again")
+
+    def check_health(self, query: Dict[str, bool] = None, diagnostic_url: str = None) -> bool:
+        """
+            Check OPA healthy. If you want check bundels or plugins, add query params for this.
+            If your diagnostic url different than default url, you can provide it.
+            ```
+            param :: query : it is the url query string. default None
+            param :: diagnostic_url : OPA diagnostic url
+
+            example:
+                print(client.check_health())
+                print(client.check_health({"bundle": True}))
+                print(client.check_health(diagnostic_url="http://localhost:8282/health"))
+                print(client.check_health(query={"bundle": True}, diagnostic_url="http://localhost:8282/health"))
+
+        """
+        if diagnostic_url:
+            url = diagnostic_url
+        else:
+            url = "{}{}:{}/{}".format(self.__schema, self.__host,
+                                      self.__port, "health")
+        if query:
+            url = self.prepare_args(url, query)
+        response = self.__session(
+            "GET", url, retries=self.retries, timeout=self.timeout)
+        if response.status == 200:
+            return True
+        return False
 
     def get_policies_list(self) -> list:
-
         """ Returns all  OPA policies in the service"""
 
         return self.__get_policies_list()
@@ -170,12 +192,12 @@ class OpaClient:
         """
         return self.__get_policies_info()
 
-    def update_opa_policy_fromstring(self, new_policy: str, endpoint: str) -> bool:
-
+    def update_opa_policy_fromstring(self, new_policy: str,
+                                     endpoint: str) -> bool:
         """ Write your rego policy with using python string type and update your OPA policies.
             ```
             param :: new_policy : is the name of your new defined  rego policy. 
-            param : endpoint : is the path of your new policy in OPA
+            param :: endpoint : is the path of your new policy in OPA
 
             example:
                 new_policy=
@@ -187,7 +209,6 @@ class OpaClient:
                         m := input.message
                         m == "world"
                     }
-                
                 client = OpaClient()
                 client.update_opa_policy_fromstring(new_policy,'exampleapi')
             ```
@@ -197,19 +218,16 @@ class OpaClient:
         return self.__update_opa_policy_fromstring(new_policy, endpoint)
 
     def update_opa_policy_fromfile(self, filepath: str, endpoint: str) -> bool:
-
         """ Write your rego policy to file and update your OPA policies. """
 
         return self.__update_opa_policy_fromfile(filepath, endpoint)
 
     def update_opa_policy_fromurl(self, url: str, endpoint: str) -> bool:
-
         """Update your OPA policies from internet."""
 
         return self.__update_opa_policy_fromurl(url, endpoint)
 
     def update_or_create_opa_data(self, new_data: dict, endpoint: str) -> bool:
-
         """ Updates existing data or create new data for policy.
             ```
             param :: new_data : name of defined data
@@ -229,12 +247,19 @@ class OpaClient:
 
         return self.__update_opa_data(new_data, endpoint)
 
-    def get_opa_raw_data(self, data_name):
-        """Returns OPA raw data in string type """
-        return self.__get_opa_raw_data(data_name)
-
-    def opa_policy_to_file(self, policy_name: str, path: Union[str,None]=None, filename: str="opa_policy.rego"):
+    def get_opa_raw_data(self, data_name: str = "", query_params: Dict[str, bool] = dict()) -> dict:
+        """ Returns OPA raw data in string type 
+            ```
+            param :: data_name : OPA data name you want get
+            param :: query_params : query params in url for more information about metrics
+            ```
+        """
+        return self.__get_opa_raw_data(data_name, query_params)
 
+    def opa_policy_to_file(self,
+                           policy_name: str,
+                           path: Union[str, None] = None,
+                           filename: str = "opa_policy.rego"):
         """ Write OPA service policy to the  file.
         ```
             param :: policy_name : name of OPA policy
@@ -251,7 +276,6 @@ class OpaClient:
         return self.__opa_policy_to_file(policy_name, path, filename)
 
     def get_opa_policy(self, policy_name: str) -> dict:
-
         """Returns full info about policy, provided OPA service """
 
         return self.__get_opa_policy(policy_name)
@@ -262,41 +286,84 @@ class OpaClient:
         return self.__delete_opa_policy(policy_name)
 
     def delete_opa_data(self, data_name: str) -> bool:
-
         """ Deletes given OPA policy data name """
 
         return self.__delete_opa_data(data_name)
 
-    def check_permission(self, input_data: dict, policy_name: str, rule_name: str) -> dict:
-
+    def check_permission(self, input_data: dict, policy_name: str,
+                         rule_name: str, query_params: Dict[str, bool] = dict()) -> dict:
         """
         ```
             params :: input_data    : data which you want check permission
                 type   :: input_data  : dict
 
-            params :: policy_name   : the name of policy resource 
+            params :: policy_name   : the name of policy resource
                 type   :: policy_name  : str
 
-            params :: rule_name   : the name included in the policy  
+            params :: rule_name   : the name included in the policy
                 type   :: rule_name  : str
+            param :: query_params : query params in url for more information about metrics
+                type :: query_params : dict
         ```
         """
 
-        return self.__check(input_data, policy_name, rule_name)
+        return self.__check(input_data, policy_name, rule_name, query_params)
 
-    def check_policy_rule(self, input_data: dict, package_path: str, rule_name: str = None) -> dict:
+    def check_policy_rule(self,
+                          input_data: dict,
+                          package_path: str,
+                          rule_name: str = None) -> dict:
         """
         Queries a package rule with the given input data
         """
 
         return self.__query(input_data, package_path, rule_name)
 
-    def __get_opa_raw_data(self, data_name: str):
-        url = self.__data_root.format(self.__root_url, data_name)
+    def ad_hoc_query(self, *, query_params: Dict[str, str] = None, body: Dict[str, str] = None):
+        """ Execute an ad-hoc query and return bindings for variables found in the query. 
+        ```
+            param :: query_params for sending query string in url
+            param :: body  for sending query in request body
 
-        response = self.__session(
-            "GET", url, headers=self.__headers, retries=2, timeout=1.5
-        )
+        ```
+        """
+
+        url = "{}{}:{}/{}/{}".format(self.__schema, self.__host,
+                                     self.__port, "v1", "query")
+        if body:
+            encoded_json = json.dumps(body).encode("utf-8")
+            response = self.__session(
+                "POST",
+                url,
+                body=encoded_json,
+                retries=self.retries,
+                timeout=self.timeout,
+            )
+        elif query_params:
+            url = self.prepare_args(url, query_params)
+            response = self.__session("GET",
+                                      url,
+                                      retries=self.retries,
+                                      timeout=self.timeout)
+        data = json.loads(response.data.decode("utf-8"))
+        if response.status == 200:
+            return data
+
+        raise QueryExecuteError(data.get("code"), data.get("message"))
+
+    def prepare_args(self, url: str, query_params: dict) -> str:
+        if query_params:
+            query_params = urlencode(query_params)
+            url = url + "?" + query_params
+        return url
+
+    def __get_opa_raw_data(self, data_name: str, query_params: Dict[str, bool]):
+        url = self.__data_root.format(self.__root_url, data_name)
+        url = self.prepare_args(url, query_params)
+        response = self.__session("GET",
+                                  url,
+                                  retries=self.retries,
+                                  timeout=self.timeout)
         code = response.status
         response = json.loads(response.data.decode("utf-8"))
         return response if code == 200 else (code, "not found")
@@ -309,9 +376,8 @@ class OpaClient:
             "PUT",
             url,
             body=encoded_json,
-            headers=self.__headers,
-            retries=2,
-            timeout=1.5,
+            retries=self.retries,
+            timeout=self.timeout,
         )
         code = response.status
         return True if code == 204 else False
@@ -321,13 +387,14 @@ class OpaClient:
         if os.path.isfile(filepath):
             with open(filepath, "r") as rf:
                 return self.__update_opa_policy_fromstring(rf.read(), endpoint)
-        
-        
-        raise FileError(f"{filepath}", "is not a file, make sure you provide a file")
 
-    def __update_opa_policy_fromstring(self, new_policy: str, endpoint: str) -> bool:
+        raise FileError(f"{filepath}",
+                        "is not a file, make sure you provide a file")
 
-        if not isinstance(new_policy,str) or not isinstance(endpoint,str):
+    def __update_opa_policy_fromstring(self, new_policy: str,
+                                       endpoint: str) -> bool:
+
+        if not isinstance(new_policy, str) or not isinstance(endpoint, str):
             raise TypeExecption(f"{new_policy} is not string type")
 
         if new_policy:
@@ -338,8 +405,8 @@ class OpaClient:
                 url,
                 body=new_policy.encode(),
                 headers=self.__headers,
-                retries=2,
-                timeout=1.5,
+                retries=self.retries,
+                timeout=self.timeout,
             )
 
             if response.status == 200:
@@ -355,9 +422,10 @@ class OpaClient:
     def __get_opa_policy(self, policy_name: str) -> dict:
         url = self.__policy_root.format(self.__root_url, policy_name)
 
-        response = self.__session(
-            "GET", url, headers=self.__headers, retries=2, timeout=1.5
-        )
+        response = self.__session("GET",
+                                  url,
+                                  retries=self.retries,
+                                  timeout=self.timeout)
         data = json.loads(response.data.decode("utf-8"))
         if response.status == 200:
 
@@ -369,7 +437,8 @@ class OpaClient:
         response = requests.get(url, headers=self.__headers)
         return self.__update_opa_policy_fromstring(response.text, endpoint)
 
-    def __opa_policy_to_file(self, policy_name: str, path: Union[str,None], filename: str) -> bool:
+    def __opa_policy_to_file(self, policy_name: str, path: Union[str, None],
+                             filename: str) -> bool:
         raw_policy = self.__get_opa_policy(policy_name)
         if isinstance(raw_policy, dict):
             try:
@@ -382,28 +451,31 @@ class OpaClient:
                         wr.write(raw_policy.get("result").get("raw").encode())
                 return True
 
-            except:
-                raise PathNotFoundError("error when write file", "path not found")
+            except:  # noqa: E722
+                raise PathNotFoundError("error when write file",
+                                        "path not found")
 
     def __delete_opa_policy(self, policy_name: str) -> bool:
         url = self.__policy_root.format(self.__root_url, policy_name)
- 
-        response = self.__session(
-            "DELETE", url, headers=self.__headers, retries=2, timeout=1.5
-        )
+
+        response = self.__session("DELETE",
+                                  url,
+                                  retries=self.retries,
+                                  timeout=self.timeout)
         data = json.loads(response.data.decode("utf-8"))
         if response.status == 200:
-
             return True
 
-        raise PolicyNotFoundError(data.get("code"), data.get("message"))
+        raise DeletePolicyError(data.get("code"), data.get("message"))
 
     def __get_policies_list(self) -> list:
         url = self.__policy_root.format(self.__root_url, "")
         temp = []
-        response = self.__session(
-            "GET", url, retries=2, timeout=1.5, headers=self.__headers
-        )
+        response = self.__session("GET",
+                                  url,
+                                  retries=self.retries,
+                                  timeout=self.timeout,
+                                  headers=self.__headers)
 
         response = json.loads(response.data.decode())
 
@@ -416,30 +488,29 @@ class OpaClient:
     def __delete_opa_data(self, data_name: str) -> bool:
         url = self.__data_root.format(self.__root_url, data_name)
 
-        response = self.__session(
-            "DELETE", url, headers=self.__headers, retries=2, timeout=1.5
-        )
+        response = self.__session("DELETE",
+                                  url,
+                                  retries=self.retries,
+                                  timeout=self.timeout)
         if response.data:
             data = json.loads(response.data.decode("utf-8"))
         if response.status == 204:
-
             return True
 
         raise DeleteDataError(data.get("code"), data.get("message"))
 
     def __get_policies_info(self) -> dict:
         url = self.__policy_root.format(self.__root_url, "")
-        policy = self.__session(
-            "GET", url, retries=2, timeout=1.0, headers=self.__headers
-        )
+        policy = self.__session("GET",
+                                url,
+                                retries=self.retries,
+                                timeout=self.timeout,
+                                headers=self.__headers)
 
         policy = json.loads(policy.data.decode())
         result = policy.get("result")
 
-        permission_url = self.__root_url
         temp_dict = {}
-        temp_policy = []
-        temp_rules = []
         for policy in result:
             temp_policy = []
             temp_rules = []
@@ -454,16 +525,22 @@ class OpaClient:
                 temp_url = permission_url
                 temp_url += "/" + rule.get("head").get("name")
                 temp_rules.append(temp_url)
-            temp_dict[policy.get("id")] = {"path": temp_policy, "rules": temp_rules}
+            temp_dict[policy.get("id")] = {
+                "path": temp_policy,
+                "rules": temp_rules
+            }
 
         return temp_dict
 
-    def __check(self, input_data: dict, policy_name: str, rule_name: str) -> dict:
+    def __check(self, input_data: dict, policy_name: str,
+                rule_name: str, query_params: Dict[str, bool]) -> dict:
         url = self.__policy_root.format(self.__root_url, policy_name)
 
-        policy = self.__session(
-            "GET", url, headers=self.__headers, retries=2, timeout=1.5
-        )
+        policy = self.__session("GET",
+                                url,
+                                headers=self.__headers,
+                                retries=self.retries,
+                                timeout=self.timeout)
         policy = json.loads(policy.data.decode("utf-8"))
         result = policy.get("result")
         find = False
@@ -480,54 +557,59 @@ class OpaClient:
                     permission_url += "/" + rule.get("head").get("name")
                     find = True
         if find:
-
             encoded_json = json.dumps(input_data).encode("utf-8")
-            response = self.__session(
-                "POST", permission_url, body=encoded_json, retries=2, timeout=1.5
-            )
+            permission_url = self.prepare_args(permission_url, query_params)
+            response = self.__session("POST",
+                                      permission_url,
+                                      body=encoded_json,
+                                      retries=self.retries,
+                                      timeout=self.timeout)
             if response.data:
                 data = json.loads(response.data.decode("utf-8"))
                 return data
 
-        raise CheckPermissionError(
-            f"{rule_name} rule not found", "path or rule name not correct"
-        )
+        raise CheckPermissionError(f"{rule_name} rule not found",
+                                   "path or rule name not correct")
 
-    def __query(self, input_data: dict, package_path: str, rule_name: str = None) -> dict:
+    def __query(self,
+                input_data: dict,
+                package_path: str,
+                rule_name: str = None) -> dict:
         if '.' in package_path:
             package_path = package_path.replace('.', '/')
         if rule_name:
             package_path = package_path + '/' + rule_name
         url = self.__data_root.format(self.__root_url, package_path)
 
-        encoded_json = json.dumps({ 'input': input_data }).encode("utf-8")
-        response = self.__session(
-            "POST", url, body=encoded_json, retries=2, timeout=1.5
-        )
+        encoded_json = json.dumps({'input': input_data}).encode("utf-8")
+        response = self.__session("POST",
+                                  url,
+                                  body=encoded_json,
+                                  retries=self.retries,
+                                  timeout=self.timeout)
         if response.data:
             data = json.loads(response.data.decode("utf-8"))
             return data
 
-        raise CheckPermissionError(
-            f"{rule_name} rule not found", "policy or rule name not correct"
-        )
+        raise CheckPermissionError(f"{rule_name} rule not found",
+                                   "policy or rule name not correct")
 
     @property
     def _host(self):
         return self.__host
-    
+
     @property
     def _port(self):
         return self.__port
-    
+
     @property
     def _version(self):
         return self.__version
-    
+
     @property
     def _root_url(self):
         return self.__root_url
-    
+
     @property
     def _schema(self):
         return self.__schema
@@ -535,11 +617,11 @@ class OpaClient:
     @property
     def _policy_root(self):
         return self.__policy_root
-    
+
     @property
     def _data_root(self):
         return self.__data_root
-    
+
     @property
     def _secure(self):
         return self.__secure
@@ -547,7 +629,7 @@ class OpaClient:
     @property
     def _ssl(self):
         return self.__ssl
-    
+
     @property
     def _cert(self):
         return self.__cert
